import multiprocessing
import pickle
import random
import uuid

from yapq import result
from yapq import serializer

class TaskRegistry:

    def __init__(self, task_list, commands_dict, result_dict, lock):
        self.task_list = task_list
        self.commands = commands_dict
        self.result_dict = result_dict
        self.lock = lock
        self.internals = task_list, commands_dict, result_dict, lock

    def put(self, job):
        self.task_list.append(serializer.encode(job))

    def get_tasks_info(self):
        return [pickle.loads(job).as_dict() for job in self.task_list]

    def find_task_index(self, uuid):
        for i, job in enumerate(self.task_list):
            job = pickle.loads(job)
            if job.uuid == uuid:
                return i
        return None

    def swap_tasks(self, left_uuid, right_uuid):
        with self.lock:
            left_idx = self.find_task_index(left_uuid)
            right_idx = self.find_task_index(right_uuid)

            print(self.get_tasks_info())
            (
                self.task_list[left_idx],
                self.task_list[right_idx],
            ) = (
                self.task_list[right_idx],
                self.task_list[left_idx],
            )
            print()
            print(self.get_tasks_info())

    def get(self, key=None):
        with self.lock:
            try:
                return serializer.decode(self.task_list.pop())
            except IndexError:
                return None

    def send_terminate_task(self):
        self.commands['terminate'] = serializer.encode(True)


    def put_result(self, value, uuid):
        result_ = result.Result()
        result_.value = value
        self.result_dict[uuid] = serializer.encode(result_)



import uuid

from yapq import result

class Job:

    def __init__(self, func, *args, **kwargs):
        self.uuid = str(uuid.uuid4())
        self.func = func
        self.args = args
        self.kwargs = kwargs

    def __call__(self):
        return self.func(*self.args, **self.kwargs)


    def as_dict(self):
        return {
            'uuid': self.uuid,
            'function': repr(self.func),
            'args': repr(self.args),
            'kwargs': repr(self.kwargs),
        }



import multiprocessing
import time

from yapq import result
from yapq import task_registry


class Worker:

    def __init__(self, task_registry):
        self.thread = multiprocessing.Process(
            target=Worker.execution_loop,
            args=(task_registry.internals)
        )
        self.thread.start()

    @staticmethod
    def execution_loop(*task_registry_internals):
        task_registry_ = task_registry.TaskRegistry(*task_registry_internals)

        while True:
            if task_registry_.commands.get('terminate'):
                break
            job = task_registry_.get()
            if not job:
                continue
            if isinstance(job, result.Result):
                continue
            value = job()
            task_registry_.put_result(value, job.uuid)

    def join(self):
        self.thread.join()



import multiprocessing
import uuid

from yapq import worker
from yapq import job
from yapq import result
from yapq import task_registry

class Yapq:

    def __init__(self):
        self.manager = multiprocessing.Manager()
        self.task_list = self.manager.list()
        self.lock = self.manager.Lock()
        self.commands_dict = self.manager.dict()
        self.result_dict = self.manager.dict()
        self.task_registry = task_registry.TaskRegistry(
            self.task_list,
            self.commands_dict,
            self.result_dict,
            self.lock,
        )

    def start(self, size=multiprocessing.cpu_count()):
        self.workers = [worker.Worker(
            self.task_registry,
            ) for _ in range(size)]

    def stop(self):
        self.task_registry.send_terminate_task()
        for w in self.workers:
            w.join()
        self.manager.shutdown()
        self.manager.join()

    def get_tasks_info(self):
        return self.task_registry.get_tasks_info()

    def swap(self, left_uuid, right_uuid):
        return self.task_registry.swap_tasks(left_uuid, right_uuid)

    def enqueue(self, func, *args, **kwargs):
        job_ = job.Job(func, *args, **kwargs)
        self.task_registry.put(job_)
        return result.Result(job_.uuid, self.result_dict)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `yapq` package."""

import pytest

from yapq import Yapq

def fn(a, b):
    return a + b


def test_yapq(monkeypatch):
    yapq = Yapq()
    res1 = yapq.enqueue(fn, 5, 3)
    res2 = yapq.enqueue(fn, 2, 3)

    yapq.start()

    assert res2.get() == 5
    assert res1.get() == 8

    yapq.stop()



import pickle


class Result:

    def __init__(self, uuid=None, result_dict=None):
        self.uuid = uuid
        self.result_dict = result_dict

    value = None

    def get(self):
        while self.uuid not in self.result_dict:
            pass
        value = pickle.loads(self.result_dict[self.uuid]).value
        del self.result_dict[self.uuid]
        return value
